import random

class Interval(object):
    '''
    This class build automatically incrementing interval objects, 
    to be used in requests timeouts
    
    Use like:
    C{interval = Interval()}
    C{t = interval()}
    C{t = interval()}

    @ivar initial:  Initial interval value, in seconds.
    @ivar maxDelay: maximun interval value produced, in seconds.
    @ivar factor:   multiplier for the next interval.
    '''
   
    
    def __init__(self, initial=2, maxDelay=1024, factor=2):
        '''Initialize interval object'''
        self.initial  = initial
        self.factor   = factor
        self.maxDelay = max(initial, maxDelay)
        self._value   = self.initial


    def __call__(self):
        '''Call the interval to produce a new delay time'''
        self._value *= self.factor
        self._value = min(self._value, self.maxDelay)
        return self._value + random.random()

class IntervalLinear(object):
    '''
    This class build automatically incrementing interval objects, 
    to be used in requests timeouts. This variant takes estimates 
    of bandwidth and Payload size and and exponential backoff
    lowering the bandwith = bandwith/factor with each new call.
    
    Use like:
    C{interval = IntervalLinear()}
    C{t = interval()}
    C{t = interval()}

    @ivar initial:  Initial interval value, in seconds.
    @ivar factor:   multiplier for the next interval.
    @ivar bandwith: estimated bandwith in bytes/sec.
    '''
    
    def __init__(self, initial=2,  factor=2, bandwith=1):
        '''Initialize interval object'''
        self.initial  = initial
        self.factor   = factor
        self.bandwith = bandwith
        self._k       = 1
        self._value   = self.initial


    def __call__(self, size):
        '''Call the interval to produce a new delay time taking into account the bandwith'''
        self._value = self.initial + (self._k*size)/self.bandwith
        self._k    *= self.factor
        return self._value + random.random()


# -*- test-case-name: mqtt.client.test.test_base -*-
# ----------------------------------------------------------------------
# Original Copyright (C) 2012 by Adam Rudd
# Modifications and redesign by Rafael Gonzalez 
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ----------------------------------------------------------------------

# +-----------------------------------------+
# | PACKET TYPES INVOLVED IN CLIENTS BY QoS |
# +-------------+-------------+-------------+
# |   QoS = 2   |   QoS = 1   |   QoS = 0   |
# +-------------+-------------+-------------+
# | CONNECT     | CONNECT     | CONNECT     | (out)(all)
# | CONNACK     | CONNACK     | CONNACK     | (in) (all)
# | DISCONNECT  | DISCONNECT  | DISCONNECT  | (out)(all)
# | PINGREQ     | PINGREQ     | PINGREQ     | (out)(all)
# | PINGRES     | PINGRES     | PINGRES     | (in) (all)
# +-------------+-------------+-------------+
# | SUBSCRIBE   | SUBSCRIBE   | SUBSCRIBE   | (out)(subscriber)
# | SUBACK      | SUBACK      | SUBACK      | (in) (subscriber)
# | UNSUBSCRIBE | UNSUBSCRIBE | UNSUBSCRIBE | (out)(subscriber)
# | UNSUBACK    | UNSUBACK    | UNSUBACK    | (in) (subscriber)
# +-------------+-------------+-------------+
# | PUBLISH     | PUBLISH     | PUBLISH     | (in/out)(subscriber/pubisher)
# | PUBREC      | PUBACK      |             | (out/in)(subscriber/publisher)
# | PUBREL      |             |             | (in/out)(subscriber/publisher)
# | PUBCOMP     |             |             | (out/in)(subscriber/publisher)
# +-------------+-------------+-------------+

# ----------------
# Standard modules
# ----------------

from random import randint

# ----------------
# Twisted  modules
# ----------------

from zope.interface import implementer
from twisted.internet.protocol import Protocol
from twisted.internet import reactor, defer, task
from twisted.logger   import Logger
from twisted.python   import failure

# -----------
# Own modules
# -----------

from ..          import v31, v311
from ..pdu       import decodeLength, DISCONNECT, PINGREQ, CONNECT, CONNACK
from ..pdu       import SUBACK, UNSUBACK, PUBLISH, PUBREL, PUBACK, PUBREC, PUBCOMP
from ..error     import ( MQTTStateError, MQTTWindowError, MQTTTimeoutError, TimeoutValueError, 
        QoSValueError, KeepaliveValueError, ClientIdValueError, ProtocolValueError, MissingTopicError,
        MissingPayloadError, MissingUserError, WindowValueError)
from .interfaces import IMQTTClientControl
from .interval   import Interval


MQTT_CONNECT_CODES = [
    "Connection Accepted",
    "Connection Refused, unacceptable protocol version",
    "Connection Refused, identifier rejected",
    "Connection Refused, Server unavailable",
    "Connection Refused, bad user name or password",
    "Connection Refused, not authorized",
]

log = Logger(namespace='mqtt')


# ---------------------------------------
# Base State Class
# ---------------------------------------

class BaseState(object):

    def __init__(self, protocol):
        self.protocol = protocol


    def connect(self, request):
        '''
        Send a CONNECT control packet.
        '''
        state = self.__class__.__name__
        return defer.fail(MQTTStateError("Unexpected connect() operation", state))


    def disconnect(self, request):
        '''
        Send a DISCONNECT control packet.
        '''
        state = self.__class__.__name__
        raise MQTTStateError("Unexpected disconnect() operation", state)


    def subscribe(self, request):
        state = self.__class__.__name__
        return defer.fail(MQTTStateError("Unexpected subscribe() operation", state))


    def unsubscribe(self, request):
        state = self.__class__.__name__
        return defer.fail(MQTTStateError("Unexpected unsubscribe() operation,", state))


    def publish(self, request):
        state = self.__class__.__name__
        return defer.fail(MQTTStateError("Unexpected publish() operation", state))


    # -------------------------------
    # Handle traffic form the network
    # -------------------------------

    # These default implementations should never be executed
    # in a well-behaved server

    def handleCONNACK(self, response):
        '''
        Handles CONNACK packet from the server
        '''
        state = self.__class__.__name__
        log.error("Unexpected {packet:7} packet received in {log_source}", packet="CONNACK")
        
    
    def handlePINGRESP(self):
        '''
        Handles PINGRESP packet from the server
        '''
        state = self.__class__.__name__
        log.error("Unexpected {packet:7} packet received in {log_source}", packet="PINGRESP")


    def handleSUBACK(self, response):
        '''
        Handles SUBACK packet from the server
        '''
        state = self.__class__.__name__
        log.error("Unexpected {packet:7} packet received in {log_source}", packet="SUBACK")


    def handleUNSUBACK(self, response):
        '''
        Decodes specific UNSUBACK data from Variable Header & Payload
        '''
        state = self.__class__.__name__
        log.error("Unexpected {packet:7} packet received in {log_source}", packet="UNSUBACK")


    def handlePUBLISH(self, response):
        '''
        Handles PUBLISH packet from the server
        '''
        state = self.__class__.__name__
        log.error("Unexpected {packet:7} packet received in {log_source}", packet="PUBLISH")


    def handlePUBACK(self, response):
        '''
        Handles PUBACK packet from the server
        '''
        state = self.__class__.__name__
        log.error("Unexpected {packet:7} packet received in {log_source}", packet="PUBACK")


    def handlePUBREC(self, response):
        '''
        Handles PUBREC packet from the server
        '''
        state = self.__class__.__name__
        log.error("Unexpected {packet:7} packet received in {log_source}", packet="PUBREC")

    def handlePUBREL(self, response):
        '''
        Handles PUBREL packet from the server
        '''
        state = self.__class__.__name__
        log.error("Unexpected {packet:7} packet received in {log_source}", packet="PUBREL")

    def handlePUBCOMP(self, response):
        '''
        Handles PUBCOMP packet from the server
        '''
        state = self.__class__.__name__
        log.error("Unexpected {packet:7} packet received in {log_source}", packet="PUBCOMP")


# ---------------------------------------
# Idle State Class
# ---------------------------------------

class IdleState(BaseState):

    def connect(self, request):
        return self.protocol.doConnect(request)

# ---------------------------------------
# Connecting State Class
# ---------------------------------------

class ConnectingState(BaseState):

    def handleCONNACK(self, response):
        self.protocol.handleCONNACK(response)

# ---------------------------------------
# Connected State Class
# ---------------------------------------

class ConnectedState(BaseState):


    def disconnect(self, request):
        '''
        Send a DISCONNECT packet.
        '''
        self.protocol.doDisconnect(request)


    def ping(self):
        '''
        Send a PINGREQ control packet.
        '''
        self.protocol.doPingRequest()


    def handlePINGRESP(self):
        '''
        Handles PINGRESP packet from the server
        '''
        self.protocol.handlePINGRESP()






# ------------------------
# MQTT Base Protocol Class
# ------------------------

@implementer(IMQTTClientControl)
class MQTTBaseProtocol(Protocol):
    '''
    Base Class, not meant to be directly instantiated.
    Handles all MQTT connection stuff
    '''
    
    # So that we can patch it in tests with Clock.callLater ...
    callLater = reactor.callLater

    packetTypes = {0x00: "null",    0x01: "CONNECT",     0x02: "CONNACK",
                   0x03: "PUBLISH", 0x04: "PUBACK",      0x05: "PUBREC",
                   0x06: "PUBREL",  0x07: "PUBCOMP",     0x08: "SUBSCRIBE",
                   0x09: "SUBACK",  0x0A: "UNSUBSCRIBE", 0x0B: "UNSUBACK",
                   0x0C: "PINGREQ", 0x0D: "PINGRESP",    0x0E: "DISCONNECT"}


    MAX_WINDOW          = 16   # Max value of in-flight PUBLISH/SUBSCRIBE/UNSUBSCRIBE
    TIMEOUT_INITIAL     = 4    # Initial tiemout for retransmissions
    TIMEOUT_MAX_INITIAL = 1024 # Maximun value for initial timeout

    def __init__(self, factory):
        self.IDLE        = IdleState(self)
        self.CONNECTING  = ConnectingState(self)
        self.CONNECTED   = ConnectedState(self)
        self.state       = self.IDLE
        self.factory     = factory
        self._initialT   = self.TIMEOUT_INITIAL # Initial timeout for retransmissions
        self._version    = v311 # default protocol version
        self._buffer     = bytearray()
        self._keepalive  = 0    # keepalive (in ms) disabled by default
        self._window     = 1    # Guarantees in-order delivery by default
        self._cleanStart = True # No session by default
        self._pingReq       = PINGREQ() 
        self._pingReq.timer = None
        self._pingReq.alarm = None
        self._pingReq.pdu   = self._pingReq.encode()    # reuses the same PDU over and over again
        self.onDisconnection = None # callback to be invoked

 # ------------------------------------------------------------------------

    def _accumulatePacket(self, data):
        self._buffer.extend(data)

        length = None

        while len(self._buffer):
            if length is None:
                # Start on a new packet

                # Haven't got enough data to start a new packet,
                # wait for some more
                if len(self._buffer) < 2:
                    break

                lenLen = 1
                # Calculate the length of the length field
                while lenLen < len(self._buffer):
                    if not self._buffer[lenLen] & 0x80:
                        break
                    lenLen += 1

                # We still haven't got all of the remaining length field
                if lenLen < len(self._buffer) and self._buffer[lenLen] & 0x80:
                    return

                length = decodeLength(self._buffer[1:])

            if len(self._buffer) >= length + lenLen + 1:
                chunk = self._buffer[:length + lenLen + 1]
                self._processPacket(chunk)
                self._buffer = self._buffer[length + lenLen + 1:]
                length = None

            else:
                break

 # ------------------------------------------------------------------------

    def _processPacket(self, packet):
        '''Generic MQTT packet decoder'''
        try:
            packet_type      = (packet[0] & 0xF0) >> 4
            packet_flags     = (packet[0] & 0x0F)
            packet_type_name = self.packetTypes[packet_type]
        except KeyError as e:
            # Invalid packet type, throw away this packet
            log.error("Invalid packet type %x" % packet_type)
            self.transport.abortConnection()
            return


        # Get the appropriate decoder function
        packetDecoder = getattr(self, "_handle%s" % packet_type_name, None)

        if packetDecoder:
            packetDecoder(packet)
        else:
            # No decoder
            log.error("Invalid packet decoder for %s" % packet_type_name)
            self.transport.abortConnection()
            return

    # -----------------------------
    # SPECIFIC MQTT PACKET DECODERS
    # -----------------------------

    def _handleCONNACK(self, packet):
        '''
        Decodes specific CONNACK data from Variable Header & Payload
        '''
        response = CONNACK()
        try:
            response.decode(packet)
        except Exception as e:
            log.debug("Exception {e!r}.", e)
            log.error("MQTT CONNACK PDU corrupt. Closing connection !")
            self.transport.abortConnection()
        else:
            self.state.handleCONNACK(response)

    # ------------------------------------------------------------------------

    def _handlePINGRESP(self, packet):
        '''
        Decodes specific PNGRESP data from Variable Header & Payload
        '''
        self.state.handlePINGRESP()           

    # ------------------------------------------------------------------------

    def _handleSUBACK(self, packet):
        '''
        Decodes specific SUBACK data from Variable Header & Payload
        '''
        response = SUBACK()
        try:
            response.decode(packet)
        except Exception as e:
            log.debug("Exception {e!r}.", e)
            log.error("MQTT SUBACK PDU corrupt. Closing connection !")
            self.transport.abortConnection()
        else:
            self.state.handleSUBACK(response)

    # ------------------------------------------------------------------------

    def _handleUNSUBACK(self, packet):
        '''
        Decodes specific UNSUBACK data from Variable Header & Payload
        '''
        response = UNSUBACK()
        try:
            response.decode(packet)
        except Exception as e:
            log.debug("Exception {e!r}.", e)
            log.error("MQTT UNSUBACK PDU corrupt. Closing connection !")
            self.transport.abortConnection()
        else:
            self.state.handleUNSUBACK(response)

    # ------------------------------------------------------------------------

    def _handlePUBLISH(self, packet):
        '''
        Decodes specific PUBLISH data from Flags, Variable Header & Payload
        '''
        response = PUBLISH()
        try:
            response.decode(packet)
        except Exception as e:
            log.debug("Exception {e!r}.", e)
            log.error("MQTT PUBLISH PDU corrupt. Closing connection !")
            self.transport.abortConnection()
        else:
            self.state.handlePUBLISH(response)

    # ------------------------------------------------------------------------

    def _handlePUBACK(self, packet):
        '''
        Decodes specific PUBACK data from Variable Header & Payload
        '''
        response = PUBACK()
        try:
            response.decode(packet)
        except Exception as e:
            log.debug("Exception {e!r}.", e)
            log.error("MQTT PUBACK PDU corrupt. Closing connection !")
            self.transport.abortConnection()
        else:
            self.state.handlePUBACK(response)

    # ------------------------------------------------------------------------

    def _handlePUBREL(self, packet):
        '''
        Decodes specific PUBREL data from Variable Header & Payload
        '''
        response = PUBREL()
        try:
            response.decode(packet)
        except Exception as e:
            log.debug("Exception {e!r}.", e)
            log.error("MQTT PUBREL PDU corrupt. Closing connection !")
            self.transport.abortConnection()
        else:
            self.state.handlePUBREL(response)

    # ------------------------------------------------------------------------

    def _handlePUBREC(self, packet):
        '''
        Decodes specific PUBREC data from Variable Header & Payload
        '''
        response = PUBREC()
        try:
            response.decode(packet)
        except Exception as e:
            log.debug("Exception {e!r}.", e)
            log.error("MQTT PUBREL PDU corrupt. Closing connection !")
            self.transport.abortConnection()
        else:
            self.state.handlePUBREC(response)

    # ------------------------------------------------------------------------

    def _handlePUBCOMP(self, packet):
        '''
        Decodes specific PUBCOMP data from Variable Header & Payload
        '''
        response = PUBCOMP()
        try:
            response.decode(packet)
        except Exception as e:
            log.debug("Exception {e!r}.", e)
            log.error("MQTT PUBCOMP PDU corrupt. Closing connection !")
            self.transport.abortConnection()
        self.state.handlePUBCOMP(response)

    # ------------------------------------------------------------------------

    # --------------------------
    # Twisted Protocol Interface
    # --------------------------

    def dataReceived(self, data):
        self._accumulatePacket(data)
    

    def connectionLost(self, reason):
        log.debug("--- Connection to MQTT Broker lost")
        if self._pingReq.timer:
            self._pingReq.timer.stop()
            self._pingReq.timer = None
        if self._pingReq.alarm:
            self._pingReq.alarm.cancel()
            self._pingReq.alarm = None
        self.doConnectionLost(reason)
        self.state = self.IDLE
        # The disconnect callback is invoked in another reactor loop cycle
        # Otherwise, the reconnection attempt happens before connection cleanup
        # which obviopusly it si not what we want.
        if self.onDisconnection:
            self.callLater(0.1, self.onDisconnection, reason)
        

    # ---------------------------------
    # IMQTTClientControl Implementation
    # ---------------------------------

    def connect(self, clientId, keepalive=0, willTopic=None,
                willMessage=None, willQoS=0, willRetain=False,
                username=None,password=None,cleanStart=True, version=v311):
        '''
        API Entry Point       
        '''
        request = CONNECT()
        request.clientId    = clientId
        request.keepalive   = keepalive
        request.willTopic   = willTopic
        request.willMessage = willMessage
        request.willQoS     = willQoS
        request.willRetain  = willRetain
        request.username    = username
        request.password    = password
        request.cleanStart  = cleanStart
        request.version     = version
        return self.state.connect(request)

    # ------------------------------------------------------------------------

    def disconnect(self):
        '''
        API Entry Point
        '''
        request = DISCONNECT()
        self.state.disconnect(request)

     # ------------------------------------------------------------------------

    def setTimeout(self, timeout):
        '''
        API Entry Point
        '''
        if not ( 1 <= timeout <= self.TIMEOUT_MAX_INITIAL ):
             raise TimeoutValueError(timeout)
        self._initialT = timeout

    # --------------------------------------------------------------------------

    def setWindowSize(self, n):
        '''
        API Entry Point
        '''
        if not (0 < n <= self.MAX_WINDOW):
            raise WindowValueError(n)
        self._window = min(n, self.MAX_WINDOW)

    # ------------------------------------------------------------------------

    def ping(self):
        '''
        Send a PINGREQ control packet
        The PINGREQ Packet is sent from a Client to the Server. 
        It can be used to:
        1. indicate to the Server that the Client is alive in the absence of 
        any other Control Packets being sent from the Client to the Server.
        2. Request that the Server responds to confirm that it is alive.
        3. Exercise the network to indicate that the Network Connection 
        is active.

         It is the responsibility of the Client to ensure that the interval 
        between Control Packets being sent does not exceed the Keep Alive value.
        In the absence of sending any other Control Packets, the Client MUST 
        send a PINGREQ Packet [MQTT-3.1.2-23].
        '''
        self.state.ping()

    # ------------------------------------------
    # Southbound interface: Network entry points
    # ------------------------------------------

    def handleCONNACK(self, response):
        '''
        Handles CONNACK packet from the server.

        The 8 bit unsigned value that represents the revision level of the 
        protocol used by the Client. The value of the Protocol Level field for
        the version 3.1.1 of the protocol is 4 (0x04). The Server MUST respond 
        to the CONNECT Packet with a CONNACK return code 0x01 (unacceptable 
        protocol level) and then disconnect the Client if the Protocol Level 
        is not supported by the Server [MQTT-3.1.2-2].

        The Keep Alive is a time interval measured in seconds. It is the 
        maximum time interval that is permitted to elapse between 
        the point at which the Client finishes transmitting one Control Packet 
        and the point it starts sending the next. A Keep Alive value of zero 
        (0) has the effect of turning off the keep alive mechanism.
        '''
        # Changes state and execute deferreds
        log.debug("<== {packet:7} (code={code} session={flags})", packet="CONNACK", code=response.resultCode, flags=response.session)
        request = self.connReq
        request.alarm.cancel()
        if response.resultCode == 0:
            self.state = self.CONNECTED
            self.mqttConnectionMade()   # before the callbacks are executed ...
            if request.keepalive != 0:
                self._pingReq.keepalive = request.keepalive
                self._pingReq.timer     = task.LoopingCall(self.ping)
                self._pingReq.timer.start(request.keepalive)
            request.deferred.callback(response.session)
        else:
            self.state = self.IDLE
            msg = MQTT_CONNECT_CODES[response.resultCode]
            request.deferred.errback(MQTTStateError(response.resultCode, msg))
        self.connReq = None     # to be garbage-collected
      
    # ------------------------------------------------------------------------
    
    def handlePINGRESP(self):
        '''
        Handles PINGRESP packet from the server
        '''
        log.debug("<== {packet:7}", packet="PINGRESP")
        self._pingReq.alarm.cancel()
        self._pingReq.alarm = None


    # ---------------------------
    # Protocol API for subclasses
    # ---------------------------

    def mqttConnectionMade(self):
        '''
        Called when a CONNACK has been received.
        Overriden in subscriber/publisher to do additional session sync
        '''
        pass

    # ---------------------------
    # State Machine API callbacks
    # ---------------------------

    def doDisconnect(self, request):
        '''
        Performs the actual work of disconnecting
        '''
        log.debug("==> {packet:7}",packet="DISCONNECT")
        self.transport.write(request.encode())
        self.transport.loseConnection()

    # ------------------------------------------------------------------------

    def doConnect(self, request):
        '''
        Performs the actual work of connecting
        '''
        def connectError():
            request.deferred.errback(MQTTTimeoutError("CONNACK"))
            request.deferred = None
            self.transport.abortConnection()            

        try:
            self._checkConnect(request)
            pdu = request.encode()
        except ValueError as e:
            return defer.fail(e)
        log.debug("==> {packet:7} (id={id} keepalive={keepalive} clean={clean})", packet="CONNECT", id=request.clientId, keepalive=request.keepalive, clean=request.cleanStart)
        self._cleanStart = request.cleanStart
        self._version    = request.version
        self.transport.write(pdu)
        # Changes state and returns deferred
        self.state = self.CONNECTING
        request.alarm = self.callLater(request.keepalive or 10, connectError)
        request.deferred = defer.Deferred()
        self.connReq = request  # keep track of this request until CONNACK or timeout
        return request.deferred

    # ------------------------------------------------------------------------

    def doPingRequest(self):
        '''
        Performs the actual work of sending PINGREQ packets
        '''
        def doPingError():
            log.warn("--- {packet:7} Timeout", packet="PINGREQ")
            self.transport.abortConnection()
        log.debug("==> {packet:7}", packet="PINGREQ")
        self.transport.write(self._pingReq.pdu)
        self._pingReq.alarm = self.callLater(self._pingReq.keepalive, doPingError)

    # ------------------------------------------------------------------------

    def doConnectionLost(self, reason):
        '''
        To bse subclassed
        '''
        pass

    # --------------
    # Helper methods
    # --------------

    def _checkConnect(self, request):
        '''
        Assert connect parameters
        '''

        if not ( 0<= request.willQoS < 3):
            raise QoSValueError("last will", request.willQoS)
        if not ( 0 <= request.keepalive <= 65535):
            raise KeepaliveValueError(request.keepalive)
        if (request.version == v31) and len(request.clientId) > 23:
            raise ClientIdValueError("Client ID exceed 23 characters", request.clientId)
        if not (request.version == v31 or request.version == v311):
            raise ProtocolValueError("Incorrect protocol version", request.version)
        if request.willMessage is not None and request.willTopic is None:
            raise MissingTopicError("Last Will Topic")
        if request.willMessage is None and request.willTopic is not None:
            raise MissingPayloadError("Last Will Message")
        if request.username is None and request.password is not None:
            raise MissingUserError()

    # ------------------------------------------------------------------------


__all__ = [MQTTBaseProtocol]


# ----------------------------------------------------------------------
# Copyright (C) 2015 by Rafael Gonzalez 
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ----------------------------------------------------------------------

class QoSValueError(ValueError):
    '''QoS value not within [0..3] range'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = '{0}: {1} {2}'.format(s, self.args[0], self.args[1])
        s = '{0}.'.format(s)
        return s

class KeepaliveValueError(ValueError):
    '''Keepalive value out of range'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = '{0}: {1}'.format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class ClientIdValueError(ValueError):
    '''Client id string length out of range'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = '{0}: {1}'.format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class ProtocolValueError(ValueError):
    '''Incorrect protocol version'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = '{0}: {1!s}'.format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class MissingTopicError(ValueError):
    '''Missing topic'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = '{0}: in {1}'.format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class MissingPayloadError(ValueError):
    '''Missing payload'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = '{0}: in {1}'.format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class MissingUserError(ValueError):
    '''Missing username in connect'''
    def __str__(self):
        s = self.__doc__
        return s


class TimeoutValueError(ValueError):
    '''Protocol timeout value out of range'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = "{0}: {1}".format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class WindowValueError(ValueError):
    '''Max. number of allowed in-flight messages out of range'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = "{0}: {1}".format(s, self.args[0])
        s = '{0}.'.format(s)
        return s


class ProfileValueError(ValueError):
    '''MQTT client profile value not supported'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = "{0}: {1}".format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class PayloadValueError(ValueError):
    '''Payload too large'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = "{0}: {1}".format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class StringValueError(ValueError):
    '''MQTT strings exceeds 65535 bytes'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = "{0}: {1}".format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class TopicTypeError(TypeError):
    '''Subscribe topic type is not a list'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = "{0}: {1!s}".format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class PayloadTypeError(TypeError):
    '''Type not allowed as payload'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = "{0}: {1!s}".format(s, self.args[0])
        s = '{0}.'.format(s)
        return s

class MQTTError(Exception):
    '''Base class for all exceptions below'''
    pass


class MQTTStateError(MQTTError):
    '''MQTT protocol state error'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = "{0}: '{1}' for {2!s}".format(s, self.args[0], self.args[1])
        s = '{0}.'.format(s)
        return s

class MQTTWindowError(MQTTError):
    '''Request exceeded maximun window size'''
    def __str__(self):
        return self.__doc__


class MQTTTimeoutError(MQTTError):
    '''Server no responding in time to expected packet'''
    def __str__(self):
        s = self.__doc__
        if self.args:
            s = "{0}: '{1}'".format(s, self.args[0])
        s = '{0}.'.format(s)
        return s



# -*- test-case-name: mqtt.client.test.test_subscriber -*-
# ----------------------------------------------------------------------
# Copyright (C) 2015 by Rafael Gonzalez 
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ----------------------------------------------------------------------

# ----------------
# Standard modules
# ----------------

# ----------------
# Twisted  modules
# ----------------

from zope.interface   import implementer
from twisted.logger   import Logger


# -----------
# Own modules
# -----------

from .interfaces import IMQTTSubscriber
from .base       import ConnectedState as BaseConnectedState
from .pubsubs    import MQTTProtocol as PubSubsMQTTProtocol


log = Logger(namespace='mqtt')

# ---------------------------------
# MQTT Client Connected State Class
# ---------------------------------

class ConnectedState(BaseConnectedState):

    def subscribe(self, request):
        return self.protocol.doSubscribe(request)

    def unsubscribe(self, request):
        return self.protocol.doUnsubscribe(request)
    
    def handleSUBACK(self, response):
        self.protocol.handleSUBACK(response)

    def handleUNSUBACK(self, response):
        self.protocol.handleUNSUBACK(response)

    def handlePUBLISH(self, response):
        self.protocol.handlePUBLISH(response)

    # QoS=2 packets
    def handlePUBREL(self, response):
        self.protocol.handlePUBREL(response)

# ------------------------
# MQTT Client Protocol Class
# ------------------------

@implementer(IMQTTSubscriber)
class MQTTProtocol(PubSubsMQTTProtocol):
    '''
    Subscriber role MQTTClient Protocol
    '''   

    def __init__(self, factory, addr):
        PubSubsMQTTProtocol.__init__(self, factory, addr)
        # patches the state machine
        self.CONNECTED = ConnectedState(self)
        

__all__ = [ "MQTTProtocol" ]



# ----------------------------------------------------------------------
# Copyright (C) 2015 by Rafael Gonzalez
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ----------------------------------------------------------------------

# ----------------
# Twisted  modules
# ----------------

import mqtt
from zope.interface import Interface, Attribute

# ============================================================================ #
#                MQTT Client Connection Control Interface                      #
# ============================================================================ #


class IMQTTClientControl(Interface):

    '''
    This interface defines operations to start,maintain and finish
    and MQTT connection, above the TCP layer
    '''

    
    def connect(clientId, keepalive=0, willTopic=None,
                willMessage=None, willQoS=0, willRetain=False, 
                username=None, password=None, cleanStart=True, version=mqtt.v311):
        '''
        Abstract
        ========

        Send a CONNECT control packet.

        Description
        ===========

        After a Network Connection is established by a Client to a Server, 
        the first Packet sent from the Client to the Server MUST be a CONNECT 
        Packet [MQTT-3.1.0-1].

        A Client can only send the CONNECT Packet once over a 
        Network Connection. The Server MUST process a second CONNECT Packet 
        sent from a Client as a protocol violation and disconnect the Client.

        If the Client does not receive a CONNACK Packet from the Server within
        a reasonable amount of time, he Client SHOULD close the Network 
        Connection. A "reasonable" amount of time depends on the type of 
        application and the communications infrastructure.

        Signature
        =========

        @param clientId: client ID for the connection (UTF-8 string)
        @param keepalive: connection keepalive period in seconds.
        @param willTopic:   last will topic  (UTF-8 string)
        @param willMessage: last will message  (UTF-8 string)
        @param willQoS:     last will qos message
        @param willRetain:  lass will retain flag.
        @param cleanStart:  session clean flag.
        @return: a Deferred whose callback will be called with a tuple
            C{returnCode, sessionFlag)} when the connection completes. 
            The Deferred errback with a C{MQTTError} exception will be called 
            if no connection ack is received from the server within a keepalive
            period. If no keepalive is used, a max of 10 seconds is used.
        '''

    def disconnect():
        '''
        Abstract
        ========

        Send a DISCONNECT packet and disconenct the transport.

        Description
        ===========

        The DISCONNECT Packet is the final Control Packet sent from 
        the Client to the Server. It indicates that the Client is 
        disconnecting cleanly. This operation is synchronous since we 
        do not expect a response from the server. A disconnect confirmation
        can be obtained through the *disconnect indication callback* below.

        Signature
        =========

        @return: Nothing.
        '''

    onDisconnection = Attribute("""
        @type onDisconnection: C{function or bounded method}
        @ivar onDisconnection: handler that will be invoked  if the Protocol loses the connection and no pending deferred remains 
        to invoke its errbacks. This is the only way for clients to be notified of such situation.
    """)

    onMqttConnectionMade = Attribute("""
        @type onMqttConnectionMade: C{function or bounded method}
        @ivar onMqttConnectionMade: handler that will be invoked before the deferreds for special cases
        (to be yet confirmed) before the connection deferred is fired.
    """)



    def setTimeout(timeout):
        '''
        Abstract
        ========

        Sets the initial timeout value for retries.

        Description
        ===========

        Set the initial timeout value for retries in PUBLISH, SUBSCRIBE, UNSUBCRIBE,
        & PUBREL control value. Retries for SUBSCRIBE, UNSUBCRIBE,
        & PUBREL will be done with exponentially backoff timeout value up to a limit. 
        Retries for PUBLISH will take into account estimated banwidth (see IPublisher)

        Signature
        =========

        @param timeout: timeout value in seconds.
        @raise ValueError: if not within [1..TIMEOUT_MAX_INITIAL]
        '''

    def setWindowSize(n):
        '''
        Abstract
        =======

        Set Acknowledge window size. 

        Description
        ===========

        Specifies the maximum number of simultaneous C{subscribe()} and
        C{unssubscribe()} requests that can be issued before waiting for
        acknowledge packets. 'n' can be limited to an internal maximun size
        (implementation defined).

        To guarantee an in-order delivery of messages for messages with QoS > 0, 
        only one ACK should be pending (n=1). 
        By default, the ack window size is n=1 unless changed by this function.

        Signature
        =========

        @param n: window size
        @raise ValueError: if not within [1..MQTTBaseProtocol.MAX_WINDOW]

        '''

# ============================================================================ #
#                      MQTT Client Subscriber Interface                        #
# ============================================================================ #

class IMQTTSubscriber(Interface):
    '''
    This interface defines the operations necessary for  a 
    pure subscriber MQTT client.
    '''

    def subscribe(topicList):
        '''
        Abstract
        ========

        Send a SUBSCRIBE control packet.
        
        Description
        ===========

        The SUBSCRIBE Packet is sent from the Client to the Server to create 
        one or more Subscriptions. Each Subscription registers a Client's 
        interest in one or more Topics.

        Signature
        =========

        @param topicList: list of tuples C{(topic, QoS)}. Each topic is
            an UTF-8 string. 0 <= QoS <= 3
        @return: a Deferred, with an extra C{msgId} attribute which you can 
            use to keep track of requests. 
            The callbacks will be invoked with a list of granted tuples (granted qos, failure flag)
            where False in the failure flag means operation Ok)
        '''

    def unsubscribe(topic):
        '''
        Abstract
        ========

        Send an UNSUBSCRIBE control packet

        Description
        ===========

        An UNSUBSCRIBE Packet is sent by the Client to the Server, 
        to unsubscribe from topics.

        Signature
        =========
        @param topic: is either a single string or a list of strings 
            [topic1, topic2, topic3,]
        @return: a Deferred, with an extra C{msgId} attribute which you can 
            use to keep track of requests. 
            The callbacks will be invoked with the msgId as parameter.
        '''


    onPublish = Attribute("""
        @type onPublish: C{function or bounded method}
        @ivar onPublish: handler that will be invoked whenever a PUBLISH message arrive.
        with parameters (topic, payload, qos, dup, retain, msgId).
    """)

    


# ============================================================================ #
#                     MQTT Client Publisher Interface                          #
# ============================================================================ #

class IMQTTPublisher(Interface):
    '''
    This interface defines the operations necessary for  a 
    pure publisher MQTT client.
    '''

    def setBandwith(bandwith, factor = 2):
        '''
        Abstract
        ========

        Set the estimate available bandwith.

         Description
        ===========

        Set the estimate available bandwith to produce timeouts proportional 
        to the payload size, according to the formula:
        T = initial + (K * payload_size)/bandwith
             where K = K*factor in each iteration
        This is useful to avoid timeouts and retransmissions in very 
        large payloads using QoS=1 and 2. 
        '''
        
    def publish(topic, message, qos=0, retain=False):
        '''

        Abstract
        ========

        Send PUBLISH control packet

        Description
        ===========
        
        Publish a message with a give QoS [0..2], returning a deferred.
        The cleanStart flag in the C{connect} API has an impact on when the
        errback is called. When cleanStart = False, 
         1) a disconenction will not cause the errback to be fired.
         2) a disconnection will not purge pending publish messages in the internal queues.

        Signature
        =========

        @param topic: an UTF-8 string describing the topic on which to publish.
        @param message: a bytearray() with the application message
        @param qos: Desired Qos to publish the message to the server [0..3].
        @param retain: Retain Flag.
        @return: a Deferred, with an extra C{msgId} attribute which you can 
            use to keep track of requests. 
            The callback is called upon successful confirm and will include
            the msgId as parameter.
        '''



# -*- test-case-name: mqtt.test.test_pdu -*-
# ----------------------------------------------------------------------
# Copyright (C) 2015 by Rafael Gonzalez 
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ----------------------------------------------------------------------


# ----------------
# Standard modules
# ----------------


# ----------------
# Twisted  modules
# ----------------

from twisted.logger import Logger

# -----------
# Own modules
# -----------

from .     import PY2, v31, v311
from .error import StringValueError, PayloadValueError, PayloadTypeError


log = Logger(namespace='mqtt')

# ---------------------------------------
# MQTT Encoding/Decoding Helper functions
# ---------------------------------------

def encodeString(string):
    '''
    Encode an UTF-8 string into MQTT format. 
    Returns a bytearray
    '''
    encoded = bytearray(2)
    encoded.extend(bytearray(string, encoding='utf-8'))
    l = len(encoded)-2
    if(l > 65535):
        raise StringValueError(l)
    encoded[0] = l >> 8
    encoded[1] = l & 0xFF
    return encoded

def decodeString(encoded):
    '''
    Decodes an UTF-8 string from an encoded MQTT bytearray.
    Returns the decoded string and renaining bytearray to be parsed
    '''
    length = encoded[0]*256 + encoded[1]
    return (encoded[2:2+length].decode('utf-8'), encoded[2+length:])


def encode16Int(value):
    '''
    Encodes a 16 bit unsigned integer into MQTT format.
    Returns a bytearray
    '''
    value      = int(value)
    encoded    = bytearray(2)
    encoded[0] = value >> 8
    encoded[1] = value & 0xFF
    return encoded

def decode16Int(encoded):
    '''
    Decodes a 16 bit unsigned integer from MQTT bytearray.
    '''
    return encoded[0]*256 + encoded[1]


def encodeLength(value):
    '''
    Encodes value into a multibyte sequence defined by MQTT protocol.
    Used to encode packet length fields.
    '''
    encoded = bytearray()
    while True:
        digit = value % 128
        value //= 128
        if value > 0:
            digit |= 128
        encoded.append(digit)
        if value <= 0:
            break
    return encoded


def decodeLength(encoded):
    '''
    Decodes a variable length value defined in the MQTT protocol.
    This value typically represents remaining field lengths
    '''
    value      = 0
    multiplier = 1
    for i in encoded:
        value += (i & 0x7F) * multiplier
        multiplier *= 0x80
        if (i & 0x80) != 0x80:
            break
    return value


# -------------------------------
# MQTT Protocol Data Units (PDUs)
# -------------------------------

class DISCONNECT(object):

    def __init__(self):
        self.encoded = None 

    def encode(self):
        '''
        Encode and store a DISCONNECT control packet.
        '''
        header    = bytearray(2)
        header[0] = 0xE0
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a DISCONNECT control packet. 
        '''
        self.encoded = packet
        

# ------------------------------------------------------------------------------

class PINGREQ(object):

    def __init__(self):
        self.encoded = None 

    def encode(self):
        '''
        Encode and store a PINGREQ control message.
        '''
        header    = bytearray(2)
        header[0] = 0xC0
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a PINGREQ control packet. 
        '''
        self.encoded = packet

# ------------------------------------------------------------------------------

# Server class Only
class PINGRES(object):

    def __init__(self):
        self.encoded = None 

    def encode(self):
        '''
        Encode and store a PINGRES control message.
        '''
        header    = bytearray(2)
        header[0] = 0xD0
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a CONNACK control packet. 
        '''
        self.encoded = packet

# ------------------------------------------------------------------------------

class CONNECT(object):

    def __init__(self):
        self.encoded    = None 
        self.clientId    = None
        self.keepalive   = None
        self.willTopic   = None
        self.willMessage = None
        self.willQoS     = None
        self.willRetain  = None
        self.username    = None
        self.password    = None
        self.cleanStart  = None
        self.version     = None

    def encode(self):
        '''
        Encode and store a CONNECT control packet. 
        @raise e: C{ValueError} if any encoded topic string exceeds 65535 bytes.
        @raise e: C{ValueError} if encoded username string exceeds 65535 bytes.
        '''
        header    = bytearray(1)
        varHeader = bytearray()
        payload   = bytearray()
        header[0] = 0x10            # packet code
        # ---- Variable header encoding section -----
        
        varHeader.extend(encodeString(self.version['tag'])) 
        varHeader.append(self.version['level']) # protocol Level
        flags =  (self.cleanStart << 1)
        if  self.willTopic is not None and self.willMessage is not None:
            flags |= 0x04 | (self.willRetain << 5) | (self.willQoS << 3)
        if self.username is not None:
            flags |= 0x80
        if self.password is not None:
            flags |= 0x40
        varHeader.append(flags)
        varHeader.extend(encode16Int(self.keepalive))
        # ------ Payload encoding section ----
        payload.extend(encodeString(self.clientId))
        if self.willTopic is not None and self.willMessage is not None:
            payload.extend(encodeString(self.willTopic))
            payload.extend(encodeString(self.willMessage))
        if self.username is not None:
             payload.extend(encodeString(self.username))
        if self.password is not None:
            payload.extend(encode16Int(len(self.password)))
            payload.extend(bytearray(self.password, encoding='ascii', errors='ignore'))
        # ---- Build the packet once all lengths are known ----
        header.extend(encodeLength(len(varHeader) + len(payload)))
        header.extend(varHeader)
        header.extend(payload)
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a CONNECT control packet. 
        '''
        self.encoded = packet
        # Strip the fixed header plus variable length field
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        # Variable Header
        version_str, packet_remaining = decodeString(packet_remaining)
        version_id = int(packet_remaining[0])
        if version_id == v31['level']:
            self.version = v31
        else:
            self.version = v311
        flags = packet_remaining[1]
        self.cleanStart = (flags & 0x02) != 0
        willFlag   = (flags & 0x04) != 0
        willQoS    = (flags >> 3) & 0x03
        willRetain = (flags & 0x20) != 0
        userFlag   = (flags & 0x80)  != 0
        passFlag   = (flags & 0x40)  != 0
        packet_remaining = packet_remaining[2:]
        self.keepalive = decode16Int(packet_remaining)
        # Payload
        packet_remaining = packet_remaining[2:]
        self.clientId, packet_remaining = decodeString(packet_remaining)
        if willFlag:
            self.willRetain = willRetain
            self.willQoS    = willQoS
            self.willTopic,  packet_remaining  = decodeString(packet_remaining)
            self.willMessage, packet_remaining = decodeString(packet_remaining)
        if userFlag:
            self.username, packet_remaining = decodeString(packet_remaining)
        if passFlag:
            l = decode16Int(packet_remaining)
            self.password = packet_remaining[2:2+l]



# ------------------------------------------------------------------------------

class CONNACK(object):

    def __init__(self):
        self.encoded   = None
        self.session    = None
        self.resultCode = None 

    def encode(self):
        '''
        Encode and store a CONNACK control packet. 
        '''
     
        header       = bytearray(1)
        varHeader    = bytearray(2)
        header[0]    = 0x20 
        varHeader[0] = self.session
        varHeader[1] = self.resultCode
        header.extend(encodeLength(len(varHeader)))
        header.extend(varHeader)
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a CONNACK control packet. 
        '''
        self.encoded = packet
        # Strip the fixed header plus variable length field
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        self.session = (packet_remaining[0] & 0x01) == 0x01 
        self.resultCode  = int(packet_remaining[1])
      
# ------------------------------------------------------------------------------

class SUBSCRIBE(object):

    def __init__(self):
        self.encoded = None
        self.topics   = None
        self.msgId    = None 

    def encode(self):
        '''
        Encode and store a SUBSCRIBE control packet. 
        @raise e: C{ValueError} if any encoded topic string exceeds 65535 bytes.
        '''
        header    = bytearray(1)
        payload   = bytearray()
        varHeader = encode16Int(self.msgId)
        header[0] = 0x82        # packet with QoS=1
        for topic in self.topics:
            payload.extend(encodeString(topic[0])) # topic name
            payload.append(topic[1])               # topic QoS
        header.extend(encodeLength(len(varHeader) + len(payload)))
        header.extend(varHeader)
        header.extend(payload)
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a SUBSCRIBE control packet. 
        '''
        self.encoded = packet
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        self.msgId   = decode16Int(packet_remaining[0:2])
        self.topics = []
        packet_remaining = packet_remaining[2:]
        while len(packet_remaining):
            topic, packet_remaining = decodeString(packet_remaining)
            qos =  int (packet_remaining[0]) & 0x03
            self.topics.append((topic,qos))
            packet_remaining = packet_remaining[1:]
        


# ------------------------------------------------------------------------------
# Server Class

class SUBACK(object):

    def __init__(self):
        self.encoded = None
        self.msgId   = None
        self.granted = None

    def encode(self):
        '''
        Encode and store a SUBACK control packet.
        '''
        header    = bytearray(1)
        payload   = bytearray()
        varHeader = encode16Int(self.msgId)
        header[0] = 0x90
        for code in self.granted:
            payload.append(code[0] | (0x80 if code[1] == True else 0x00))
        header.extend(encodeLength(len(varHeader) + len(payload)))
        header.extend(varHeader)
        header.extend(payload)
        self.encoded = header
        return str(header) if PY2 else bytes(header)


    def decode(self, packet):
        '''
        Decode a SUBACK control packet. 
        '''
        self.encoded = packet
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        self.msgId   = decode16Int(packet_remaining)
        # Make a sequence of tuples of (GrantedQoS, Failure Flag)
        self.granted = [ (byte & 0x7F, byte & 0x80 == 0x80) 
                    for byte in packet_remaining[2:] ]




# ------------------------------------------------------------------------------

class UNSUBSCRIBE(object):

    def __init__(self):
        self.encoded = None
        self.msgId   = None
        self.topics  = None

    def encode(self):
        '''
        Encode and store an UNSUBCRIBE control packet
        @raise e: C{ValueError} if any encoded topic string exceeds 65535 bytes
        '''
        header    = bytearray(1)
        payload   = bytearray()
        varHeader = encode16Int(self.msgId)
        header[0] = 0xA2    # packet with QoS=1
        for topic in self.topics:
            payload.extend(encodeString(topic)) # topic name
        header.extend(encodeLength(len(varHeader) + len(payload)))
        header.extend(varHeader)
        header.extend(payload)
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a UNSUBACK control packet. 
        '''
        self.encoded = packet
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        self.msgId   = decode16Int(packet_remaining[0:2])
        self.topics = []
        packet_remaining = packet_remaining[2:]
        while len(packet_remaining):
            l = decode16Int(packet_remaining[0:2])
            topic = packet_remaining[2:2+l].decode(encoding='utf-8')
            self.topics.append(topic)
            packet_remaining = packet_remaining[2+l:]

# ------------------------------------------------------------------------------

# Server PDU
class UNSUBACK(object):

    def __init__(self):
        self.encoded = None
        self.msgId   = None

    def encode(self):
        '''
        Encode and store an UNSUBACK control packet
        '''
        header    = bytearray(1)
        varHeader = encode16Int(self.msgId)
        header[0] = 0xB0 
        header.extend(encodeLength(len(varHeader)))
        header.extend(varHeader)
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a UNSUBACK control packet. 
        '''
        self.encoded = packet
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        self.msgId   = decode16Int(packet_remaining)


# ------------------------------------------------------------------------------

class PUBLISH(object):

    def __init__(self):
        self.encoded = None
        self.qos     = None
        self.dup     = None
        self.retain  = None
        self.topic   = None
        self.msgId   = None
        self.payload = None

    def encode(self):
        '''
        Encode and store a PUBLISH control packet.
        @raise e: C{ValueError} if encoded topic string exceeds 65535 bytes.
        @raise e: C{ValueError} if encoded packet size exceeds 268435455 bytes.
        @raise e: C{TypeError} if C{data} is not a string, bytearray, int, boolean or float.
        '''
        header    = bytearray(1)
        varHeader = bytearray()
        payload   = bytearray()

        if self.qos:
            header[0] = 0x30 | self.retain | (self.qos << 1) | (self.dup << 3)
            varHeader.extend(encodeString(self.topic)) # topic name
            varHeader.extend(encode16Int(self.msgId))  # msgId should not be None
        else:
            header[0] = 0x30 | self.retain
            varHeader.extend(encodeString(self.topic)) # topic name
        if isinstance(self.payload, bytearray):
            payload.extend(self.payload)
        elif isinstance(self.payload, str):
            payload.extend(bytearray(self.payload, encoding='utf-8'))
        else:
            raise PayloadTypeError(type(self.payload))
        totalLen = len(varHeader) + len(payload)
        if totalLen > 268435455:
            raise PayloadValueError(totalLen)
        header.extend(encodeLength(totalLen))
        header.extend(varHeader)
        header.extend(payload)
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a PUBLISH control packet. 
        '''
        self.encoded = packet
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        self.dup    = (packet[0] & 0x08) == 0x08
        self.qos    = (packet[0] & 0x06) >> 1
        self.retain = (packet[0] & 0x01) == 0x01
        self.topic, _  = decodeString(packet_remaining)
        topicLen       = decode16Int(packet_remaining)
        if self.qos:
            self.msgId = decode16Int( packet_remaining[topicLen+2:topicLen+4] )
            self.payload =  packet_remaining[topicLen+4:]
        else:
            self.msgId = None
            self.payload = packet_remaining[topicLen+2:] # payload is a bytearray
        

# ------------------------------------------------------------------------------

class PUBACK(object):

    def __init__(self):
        self.encoded = None
        self.msgId   = None

    def encode(self):
        '''
        Encode and store a PUBACK control packet
        '''
        header    = bytearray(1)
        varHeader = encode16Int(self.msgId)
        header[0] = 0x40 
        header.extend(encodeLength(len(varHeader)))
        header.extend(varHeader)
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a PUBACK control packet. 
        '''
        self.encoded = packet
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        self.msgId = decode16Int(packet_remaining)


# ------------------------------------------------------------------------------

class PUBREC(object):
   
    def __init__(self):
        self.encoded = None
        self.msgId   = None

    def encode(self):
        '''
        Encode and store a PUBREC control packet
        '''
        header    = bytearray(1)
        varHeader = encode16Int(self.msgId)
        header[0] = 0x50 
        header.extend(encodeLength(len(varHeader)))
        header.extend(varHeader)
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a PUBREC control packet. 
        '''
        self.encoded = packet
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        self.msgId = decode16Int(packet_remaining)


# ------------------------------------------------------------------------------

class PUBREL(object):
   
    def __init__(self):
        self.encoded = None
        self.msgId   = None
        self.dup     = None

    def encode(self):
        '''
        Encode and store a PUBREL control packet
        '''
        header    = bytearray(1)
        varHeader = encode16Int(self.msgId)
        header[0] = 0x62    # packet with QoS=1
        header.extend(encodeLength(len(varHeader)))
        header.extend(varHeader)
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a PUBREL control packet. 
        '''
        self.encoded = packet
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        self.msgId  = decode16Int(packet_remaining)
        self.dup = (packet[0] & 0x08) == 0x08


# ------------------------------------------------------------------------------

class PUBCOMP(object):
   
    def __init__(self):
        self.encoded = None
        self.msgId    = None
        
    def encode(self):
        '''
        Encode and store a PUBCOMP control packet
        '''
        header    = bytearray(1)
        varHeader = encode16Int(self.msgId)
        header[0] = 0x72 
        header.extend(encodeLength(len(varHeader)))
        header.extend(varHeader)
        self.encoded = header
        return str(header) if PY2 else bytes(header)

    def decode(self, packet):
        '''
        Decode a PUBCOMP control packet. 
        '''
        self.encoded = packet
        lenLen = 1
        while packet[lenLen] & 0x80:
            lenLen += 1
        packet_remaining = packet[lenLen+1:]
        self.msgId   = decode16Int(packet_remaining)

# ------------------------------------------------------------------------------

__all__ = [
    decodeLength,
    CONNECT,
    CONNACK,
    DISCONNECT,
    PINGREQ,
    PINGRES,
    SUBSCRIBE,
    SUBACK,
    UNSUBSCRIBE,
    UNSUBACK,
    PUBLISH,
    PUBACK,
    PUBREC,
    PUBREL,
    PUBCOMP,
]



# ----------------------------------------------------------------------
# Copyright (C) 2015 by Rafael Gonzalez
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ----------------------------------------------------------------------

# ----------------
# Standard modules
# ----------------

from collections import deque

# ----------------
# Twisted  modules
# ----------------

from zope.interface   import implementer
from twisted.internet import defer, reactor, error
from twisted.logger   import Logger

# -----------
# Own modules
# -----------

from ..          import v31, PY2
from ..error     import MQTTWindowError, QoSValueError, TopicTypeError
from ..pdu       import SUBSCRIBE, UNSUBSCRIBE, PUBACK, PUBREC, PUBCOMP, PUBLISH, PUBREL
from .interfaces import IMQTTSubscriber, IMQTTPublisher
from .interval   import Interval, IntervalLinear
from .base       import MQTTBaseProtocol, IdleState as BaseIdleState, ConnectingState as BaseConnectingState, ConnectedState as BaseConnectedState


log = Logger(namespace='mqtt')


class MQTTSessionCleared(Exception):
    '''MQTT persitent session cleared and message could not be published.'''
    def __str__(self):
        return self.__doc__

# ---------------------------------------------
# MQTT Client Idle State Class (for subscriber)
# ---------------------------------------------

class IdleState(BaseIdleState):
    pass


# --------------------------------------------------
# MQTT Client Connecting State Class (for publisher)
# --------------------------------------------------

class ConnectingState(BaseConnectingState):

    def handleCONNACK(self, response):
        self.protocol.handleCONNACK(response)

    # The standard allows publishing data without waiting for CONNACK
    def publish(self, request):
        return self.protocol.doPublish(request)

# ---------------------------------
# MQTT Client Connected State Class
# ---------------------------------

class ConnectedState(BaseConnectedState):

    def publish(self, request):
        return self.protocol.doPublish(request)

    def subscribe(self, request):
        return self.protocol.doSubscribe(request)

    def unsubscribe(self, request):
        return self.protocol.doUnsubscribe(request)
    
    def handleSUBACK(self, response):
        self.protocol.handleSUBACK(response)

    def handleUNSUBACK(self, response):
        self.protocol.handleUNSUBACK(response)

    def handlePUBLISH(self, response):
        self.protocol.handlePUBLISH(response)

    # QoS=1 packets forwarded to subscriber
    def handlePUBACK(self, response):
        self.protocol.handlePUBACK(response)

    # QoS=2 packets forwarded to publisher
    def handlePUBREC(self, response):
        self.protocol.handlePUBREC(response)

    # QoS=2 packets forwarded to subscriber
    def handlePUBREL(self, response):
        self.protocol.handlePUBREL(response)

    # QoS=2 packets forwarded to publisher
    def handlePUBCOMP(self, response):
        self.protocol.handlePUBCOMP(response)

# --------------------------
# MQTT Client Protocol Class
# --------------------------

@implementer(IMQTTSubscriber, IMQTTPublisher)
class MQTTProtocol(MQTTBaseProtocol):
    '''
    MQTTClient publish/subscribe Protocol
    '''

    DEFAULT_BANDWITH = 10000
    DEFAULT_FACTOR   = 2

    def __init__(self, factory, addr):
        MQTTBaseProtocol.__init__(self, factory) 
        # patches and reparent the state machine
        self.IDLE          = IdleState(self)
        self.CONNECTING    = ConnectingState(self)
        self.CONNECTED     = ConnectedState(self)
        self.state         = self.IDLE
        # Copies addr
        self.addr          = addr
        # Estimated bandwith in bytes/sec for PUBLISH PDUs
        self._bandwith     =  self.DEFAULT_BANDWITH
        self._factor       =  self.DEFAULT_FACTOR
        # additional, per-connection subscriber state
        self.onPublish   = None
        # a callback  when CONNACK packet is received
        self.onMqttConnectionMade = None  
      
       
    # -----------------------------
    # IMQTTPublisher Implementation
    # -----------------------------
  
    def setBandwith(self, bandwith, factor=2):
        if bandwith <= 0:
            raise ValueError("Bandwith should be a positive number")
        if factor <= 0:
            raise ValueError("Factor should be a positive number")
        self._bandwith = bandwith
        self._factor   = factor

    
    def publish(self, topic, message, qos=0, retain=False):
        '''
        API entry point.
        '''
        request = PUBLISH()
        request.qos     = qos
        request.topic   = topic
        request.payload = message
        request.retain  = retain
        request.dup     = False
        return self.state.publish(request)


    # ---------------------------------
    # IMQTTSubscriber Implementation
    # ---------------------------------

    def subscribe(self, topics, qos=0):
        '''
        API entry point.
        '''
        request = SUBSCRIBE()
        request.topics = topics
        request.qos    = qos
        return self.state.subscribe(request)

    # --------------------------------------------------------------------------

    def unsubscribe(self, topics):
        '''
        API entry point.
        '''
        request = UNSUBSCRIBE()
        request.topics = topics
        return self.state.unsubscribe(request)


    # ------------------------------------------
    # Southbound interface: Network entry points
    # ------------------------------------------

    # handleSUBACK(), handleUNSUBACK() handlePUBLISH() & handlePUBREL() for subscribers
    def handleSUBACK(self, response):
        '''
        Handle SUBACK control packet received.
        '''
        try:
            request = self.factory.windowSubscribe[self.addr][response.msgId]
        except KeyError as e:
            log.debug("<== {packet:7} (id={response.msgId:04x}) already handled" , packet="SUBACK",  response=response)
        else:    
            log.debug("<== {packet:7} (id={response.msgId:04x})" , packet="SUBACK",  response=response)
            request = self.factory.windowSubscribe[self.addr][response.msgId]
            del self.factory.windowSubscribe[self.addr][response.msgId]
            request.alarm.cancel()
            request.deferred.callback(response.granted)
       
    # --------------------------------------------------------------------------

    def handleUNSUBACK(self, response):
        '''
        Handle UNSUBACK control packet received.
        '''
        try:
            request = self.factory.windowUnsubscribe[self.addr][response.msgId]
        except KeyError as e:
            log.debug("<== {packet:7} (id={response.msgId:04x}) already handled" , packet="UNSUBACK",  response=response)
        else:
            log.debug("<== {packet:7} (id={response.msgId:04x})" , packet="UNSUBACK",  response=response)
            request = self.factory.windowUnsubscribe[self.addr][response.msgId]
            del self.factory.windowUnsubscribe[self.addr][response.msgId]
            request.alarm.cancel()
            request.deferred.callback(response.msgId)

    # --------------------------------------------------------------------------

    def handlePUBLISH(self, response):
        '''
        Handle PUBLISH control packet received.
        '''
        if  response.qos == 0:
            log.debug("==> {packet:7} (id={response.msgId} qos={response.qos} dup={response.dup} retain={response.retain} topic={response.topic})" , packet="PUBLISH", response=response)
            self._deliver(response)
        elif response.qos == 1:
            log.debug("==> {packet:7} (id={response.msgId:04x} qos={response.qos} dup={response.dup} retain={response.retain} topic={response.topic})" , packet="PUBLISH", response=response)
            reply = PUBACK()
            reply.msgId = response.msgId
            log.debug("<== {packet:7} (id={response.msgId:04x})" , packet="PUBACK", response=response)
            self.transport.write(reply.encode())
            self._deliver(response)
        elif response.qos == 2:
            log.debug("==> {packet:7} (id={response.msgId:04x} qos={response.qos} dup={response.dup} retain={response.retain} topic={response.topic})" , packet="PUBLISH", response=response)
            self.factory.windowPubRx[self.addr][response.msgId] = response
            reply = PUBREC()
            reply.msgId = response.msgId
            log.debug("<== {packet:7} (id={response.msgId:04x})" , packet="PUBREC", response=response)
            self.transport.write(reply.encode())

    # --------------------------------------------------------------------------

    def handlePUBREL(self, response):
        '''
        Handle PUBREL control packet received.
        '''
        try:
            msg = self.factory.windowPubRx[self.addr][response.msgId]
        except KeyError as e:
            log.debug("==> {packet:7}(id={response.msgId:04x} dup={response.dup}) already handled" , packet="PUBREL", response=response)
        else:
            log.debug("==> {packet:7}(id={response.msgId:04x} dup={response.dup})" , packet="PUBREL", response=response)
            del self.factory.windowPubRx[self.addr][response.msgId]
            self._deliver(msg)
            reply = PUBCOMP()
            reply.msgId = response.msgId
            reply.encode()
            log.debug("<== {packet:7} (id={response.msgId:04x})" , packet="PUBCOMP", response=response)
            self.transport.write(reply.encode())


    # --------------------------------------------------------------------------

    # handlePUBACK(), handlePUBREC() & handlePUBCOMP() are for publisher

    def handlePUBACK(self, response):
        '''
        Handle PUBACK control packet received (QoS=1).
        '''
        # so:  response.msgId == windowPublish[self.addr][0].msgId
        try:
             request = self.factory.windowPublish[self.addr][response.msgId]
        except KeyError as e:
            log.debug("<== {packet:7} (id={response.msgId:04x}) already handled", packet="PUBACK", response=response)
        else:
            log.debug("<== {packet:7} (id={response.msgId:04x})", packet="PUBACK", response=response)
            request.alarm.cancel()
            request.deferred.callback(request.msgId)
            del self.factory.windowPublish[self.addr][response.msgId]
            self._refillPublish(dup=False)

    # --------------------------------------------------------------------------

    def handlePUBREC(self, response):
        '''
        Handle PUBREC control packet received (QoS=2).
        '''
        # so:  response.msgId == windowPublish[self.addr][0].msgId
        try:
            request = self.factory.windowPublish[self.addr][response.msgId]
        except KeyError as e:
            log.debug("<== {packet:7} (id={response.msgId:04x}) already handled", packet="PUBREC", response=response)
        else:
            log.debug("<== {packet:7} (id={response.msgId:04x})", packet="PUBREC", response=response)
            request.alarm.cancel()
            del self.factory.windowPublish[self.addr][response.msgId]
            reply = PUBREL()
            reply.msgId = response.msgId
            reply.interval = Interval(initial=self._initialT)
            reply.deferred = request.deferred       # Transfer the deferred to PUBREL
            reply.retries  = request.retries        # and the retry count
            reply.encode()
            self.factory.windowPubRelease[self.addr][reply.msgId] = reply
            self._retryRelease(reply, False)


    # --------------------------------------------------------------------------

    def handlePUBCOMP(self, response):
        '''
        Handle PUBCOMP control packet received (QoS=2).
        '''
        # Same comment as PUBACK
        try:
            reply = self.factory.windowPubRelease[self.addr][response.msgId]
        except KeyError as e:
            log.debug("<== {packet:7} (id={response.msgId:04x}) already handled", packet="PUBCOMP", response=response)
        else: 
            log.debug("<== {packet:7} (id={response.msgId:04x})", packet="PUBCOMP", response=response)
            reply.alarm.cancel()
            reply.deferred.callback(reply.msgId)
            del self.factory.windowPubRelease[self.addr][reply.msgId]
            self._refillPublish(dup=False)


    # ---------------------------
    # Protocol API for subclasses
    # ---------------------------

    def mqttConnectionMade(self):
        '''
        Called when a CONNACK has been received (publisher only).
        '''
        if self._cleanStart:
            self._purgeSession()
        else:
            self._syncSession()
        if self.onMqttConnectionMade:
            self.onMqttConnectionMade()

    # ---------------------------
    # State Machine API callbacks
    # ---------------------------

    def doSubscribe(self, request):
        '''
        Send a SUBSCRIBE control packet.
        '''
        
        if isinstance(request.topics, str):
            request.topics = [(request.topics, request.qos)] 
        elif isinstance(request.topics, tuple):
            request.topics = [(request.topics[0], request.topics[1])] 
        try:
            self._checkSubscribe(request)
            request.msgId = self.factory.makeId()
            request.encode()
        except Exception as e:
            return defer.fail(e)
        request.interval = Interval(initial=self._initialT)
        request.deferred = defer.Deferred()
        request.deferred.msgId = request.msgId
        self.factory.windowSubscribe[self.addr][request.msgId] = request
        self._retrySubscribe(request, False)
        return  request.deferred 

    # --------------------------------------------------------------------------

    def doUnsubscribe(self, request):
        '''
        Send an UNSUBSCRIBE control packet
        '''
        request.msgId = self.factory.makeId()
        if isinstance(request.topics, str):
            request.topics = [request.topics]
        try:
            self._checkUnsubscribe(request)
            request.msgId = self.factory.makeId()
            request.encode() 
        except Exception as e:
            return defer.fail(e)
        request.interval = Interval(initial=self._initialT)
        request.deferred = defer.Deferred()
        request.deferred.msgId = request.msgId
        self.factory.windowUnsubscribe[self.addr][request.msgId] = request
        self._retryUnsubscribe(request, dup=False)
        return  request.deferred

    # --------------------------------------------------------------------------

    def doSetPublishHandler(self, callback):
        '''
        Stores the publish callback
        '''
        self.onPublish = callback

    # --------------------------------------------------------------------------

    def doPublish(self, request):
        '''
        Send PUBLISH control packet
        '''

        try:
            self._checkPublish(request)
        except Exception as e:
            return defer.fail(e)

        if request.qos == 0:
            request.msgId    = None
            request.deferred = defer.succeed(None)
            request.interval = None
        else:
            request.msgId    = self.factory.makeId()
            request.deferred = defer.Deferred()
            request.interval = IntervalLinear(initial=self._initialT, 
                                              bandwith=self._bandwith, 
                                              factor=self._factor)
            request.retries  = 0
        try:
            request.encode()
        except Exception as e:
            return defer.fail(e)

        self.factory.queuePublishTx[self.addr].append(request)
        request.deferred.msgId = request.msgId
        self._refillPublish(dup=False)
        return  request.deferred 


    # --------------------------
    # Helper methods (subscriber)
    # ---------------------------

    def _retrySubscribe(self, request, dup):
        '''
        Transmit/Retransmit SUBSCRIBE packet
        '''
        if self._version == v31:
            request.encoded[0] |=  (dup << 3)   # set the dup flag
        interval = request.interval() + 0.25*len(self.factory.windowSubscribe[self.addr])
        request.alarm = self.callLater(interval, self._subscribeError, request)
        log.debug("==> {packet:7} (id={request.msgId:04x} dup={dup})", packet="SUBSCRIBE", request=request, dup=dup)
        self.transport.write(str(request.encoded) if PY2 else bytes(request.encoded))

    # --------------------------------------------------------------------------

    def _retryUnsubscribe(self, request, dup):
        '''
        Transmit/Retransmit UNSUBSCRIBE packet
        '''
        if self._version == v31:
            request.encoded[0] |=  (dup << 3)   # set the dup flag
        interval = request.interval() + 0.25*len(self.factory.windowUnsubscribe[self.addr])
        request.alarm = self.callLater(interval, self._unsubscribeError, request)
        log.debug("==> {packet:7} (id={request.msgId:04x} dup={dup})", packet="UNSUBSCRIBE", request=request, dup=dup)
        self.transport.write(str(request.encoded) if PY2 else bytes(request.encoded))

    # --------------------------------------------------------------------------

    def _deliver(self, pdu):
        '''Deliver the message to the client if registered a callback'''
        if self.onPublish:
            self.onPublish(pdu.topic, pdu.payload, pdu.qos, pdu.dup, pdu.retain, pdu.msgId)

    # --------------------------------------------------------------------------

    def _subscribeError(self, request):
        '''
        Handle lack of SUBACK
        '''
        log.error("{packet:7} (id={request.msgId:04x}) {timeout}, retransmitting", packet="SUBSCRIBE", request=request,  timeout="timeout")
        self._retrySubscribe(request,  dup=True)

    # --------------------------------------------------------------------------

    def _unsubscribeError(self, request):
        '''
        Handle ack of UNSUBACK packet
        '''
        log.error("{packet:7} (id={request.msgId:04x}) {timeout}, retransmitting", packet="UNSUBSCRIBE", request=request,  timeout="timeout")
        self.reUnubscribe(request,  dup=True)

    # --------------------------------------------------------------------------

    def _checkSubscribe(self, request):
        '''
        Assert subscribe parameters
        '''
        if len(self.factory.windowSubscribe[self.addr]) == self._window:
            raise MQTTWindowError("subscription requests exceeded limit", self._window)
        if not isinstance(request.topics, list):
            raise TopicTypeError(type(topic))
        for (topic, qos) in request.topics:
            if not ( 0<= qos < 3):
                raise QoSValueError("subscribe", qos)

    # --------------------------------------------------------------------------

    def _checkUnsubscribe(self, request):
        '''
        Assert unsubscribe parameters
        '''
        if len(self.factory.windowUnsubscribe[self.addr]) == self._window:
            raise MQTTWindowError("unsubscription requests exceeded limit", self._window)
        if not isinstance(request.topics, list):
            raise TopicTypeError(type(topic))

    # --------------------------
    # Helper methods (publisher)
    # --------------------------

    def _refillPublish(self, dup):
        '''
        Refills the Publisher transmission window from the queue 
        '''
        cnx = self.addr
        N = min(self._window - len(self.factory.windowPublish[cnx]), len(self.factory.queuePublishTx[cnx]))
        for i in range(0,N):
            request = self.factory.queuePublishTx[cnx].popleft()
            if request.msgId:   # only form QoS 1 & 2
                self.factory.windowPublish[cnx][request.msgId] = request
            self._retryPublish(request, dup)


    def _retryPublish(self, request, dup):
        '''
        Transmit/Retransmit one PUBLISH packet 
        '''
        request.encoded[0] |=  (dup << 3)   # set the dup flag
        request.dup = dup
        if request.interval:    # Handle timeouts for QoS 1 and 2
            request.alarm = self.callLater(request.interval(len(request.encoded)), self._publishError, request)
        if request.msgId is None:
            log.debug("==> {packet:7} (id={request.msgId} qos={request.qos} dup={dup} retain={request.retain} topic={request.topic})", packet="PUBLISH", request=request, dup=dup)
        else:
            log.debug("==> {packet:7} (id={request.msgId:04x} qos={request.qos} dup={dup} retain={request.retain} topic={request.topic})", packet="PUBLISH", request=request, dup=dup)
        self.transport.write(str(request.encoded) if PY2 else bytes(request.encoded))

    # --------------------------------------------------------------------------

    def _retryRelease(self, reply, dup):
        '''
        Transmit/Retransmit PUBREL packet 
        '''
        if self._version == v31:
            reply.encoded[0] |=  (dup << 3)   # set the dup flag
            reply.dup = dup
        reply.alarm = self.callLater(reply.interval(), self._pubrelError, reply)
        log.debug("==> {packet:7} (id={reply.msgId:04x} dup={dup})", packet="PUBREL", reply=reply, dup=dup)
        self.transport.write(str(reply.encoded) if PY2 else bytes(reply.encoded))

    # --------------------------------------------------------------------------


    # According to QoS = 1 we should never give up _retryPublish as long as 
    # we are connected to a server. So there is no retry count.

    def _publishError(self, request):
        '''
        Handle the absence of PUBACK / PUBREC 
        '''
        request.retries += 1
        log.error("{packet:7} (id={request.msgId:04x} qos={request.qos}) {timeout}, _retryPublish({request.retries})", packet="PUBREC/PUBACK", request=request, timeout="timeout")
        self._retryPublish(request, dup=True)

    # --------------------------------------------------------------------------

    def _pubrelError(self, reply):
        '''
        Handle the absence of PUBCOMP 
        '''
        log.error("{packet:7} (id={request.msgId:04x}) {timeout}, _retryPublish", packet="PUBCOMP", timeout="timeout")
        self._retryRelease(reply, dup=True)

    # --------------------------------------------------------------------------

    def _checkPublish(self, request):
        '''
        Assert publish parameters
        '''
        if not ( 0<= request.qos < 3):
            raise QoSValueError("publish()",request.qos)
    
    # --------------------------------------------------------------------------

    def _syncSession(self):
        '''
        Tries to restore the session state upon a new MQTT connection made (publisher)
        '''
        #log.debug("{event}", event="Sync Persistent Session")
        for _, reply in self.factory.windowPubRelease[self.addr].items():
            self._retryRelease(reply, dup=True)
        for _, request in self.factory.windowPublish[self.addr].items():
            self._retryPublish(request, dup=True)

    # --------------------------------------------------------------------------

    def _purgeSession(self):
        '''
        Purges the persistent state in the client 
        '''
        #log.debug("{event}", event="Clean Persistent Session")
        for k in self.factory.windowPublish[self.addr].keys():
            request = self.factory.windowPublish[self.addr][k]
            del self.factory.windowPublish[self.addr][k]
            request.deferred.errback(MQTTSessionCleared)

        for k in self.factory.windowPubRelease[self.addr].keys():
            request = self.factory.windowPubRelease[self.addr][k]
            del self.factory.windowPubRelease[self.addr][k]
            request.deferred.errback(MQTTSessionCleared)


    # -------------------------------------
    # Helper methods (publisher/subscriber)
    # -------------------------------------

    def doConnectionLost(self, reason):
        '''
        Additional connection lost clean up.
        '''
       
        # Cancel Alarms first
        for _, request in self.factory.windowSubscribe[self.addr].items():
            if request.alarm is not None:
                request.alarm.cancel()
                request.alarm = None
        for _, request in self.factory.windowUnsubscribe[self.addr].items():
            if request.alarm is not None:
                request.alarm.cancel()
                request.alarm = None
        for _, request in self.factory.windowPublish[self.addr].items():
            if request.alarm is not None:
                request.alarm.cancel()
                request.alarm = None
        for _, request in self.factory.windowPubRelease[self.addr].items():
            if request.alarm is not None:
                request.alarm.cancel()
                request.alarm = None
        # Then, invoke errbacks anyway if we do not persist state
        if self._cleanStart:
            for k in self.factory.windowSubscribe[self.addr].keys():
                request = self.factory.windowSubscribe[self.addr][k]
                del self.factory.windowSubscribe[self.addr][k]
                request.deferred.errback(reason)
            for k in self.factory.windowUnsubscribe[self.addr].keys():
                request = self.factory.windowUnsubscribe[self.addr][k]
                del self.factory.windowUnsubscribe[self.addr][k]
                request.deferred.errback(reason)
            for k in self.factory.windowPubRelease[self.addr].keys():
                request = self.factory.windowPubRelease[self.addr][k]
                del self.factory.windowPubRelease[self.addr][k]
                request.deferred.errback(reason)                
            for k in self.factory.windowPublish[self.addr].keys():
                request = self.factory.windowPublish[self.addr][k]
                del self.factory.windowPublish[self.addr][k]
                request.deferred.errback(reason)   

__all__ = [ "MQTTProtocol" ]



# -*- test-case-name: mqtt.client.test.test_factory -*-
# ----------------------------------------------------------------------
# Copyright (C) 2015 by Rafael Gonzalez 
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ----------------------------------------------------------------------

# ----------------
# Standard modules
# ----------------

from collections import deque

# ----------------
# Twisted  modules
# ----------------

# ReconnectingClientFactory is becoming obsolete
# since applications now have ClientService and its retryPolicy parameter
# See chapter "Getting Connected with Endpoints" in the Twisted manual
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.logger import Logger

# -----------
# Own modules
# -----------

from ..      import __version__
from ..error import ProfileValueError

log = Logger(namespace='mqtt')

class MQTTFactory(ReconnectingClientFactory):


    SUBSCRIBER = 0x1
    PUBLISHER  = 0x2


    def __init__(self, profile):
        self.profile  = profile
        self.factor   = 2
        self.maxDelay = 2*3600
        # Packet Id generator
        self.id       = 0
        self.queuePublishTx    = {} # PUBLISH messages waiting before being transmitted
        self.windowPublish     = {} # PUBLISH messages window waiting for PUBREC/PUBACK
        self.windowPubRelease  = {} # PUBREL  messages (qos=2) window waiting for PUBCOMP (publisher)
        self.windowPubRx       = {} # PUBLISH messages (qos=2) window waiting for PUBREL (subscriber side)
        self.windowSubscribe   = {} # SUBSCRIBE messages window, waiting fr SUBACK
        self.windowUnsubscribe = {} # UNSUBSCRIBE messages window, waiting fr UNSUBACK

        log.info("MQTT Client library version {version}", version=__version__)
    

    def buildProtocol(self, addr):
        if   self.profile == self.SUBSCRIBER:
            from mqtt.client.subscriber import MQTTProtocol
        elif self.profile == self.PUBLISHER:
            from mqtt.client.publisher import MQTTProtocol
        elif self.profile == (self.SUBSCRIBER | self.PUBLISHER):
            from mqtt.client.pubsubs import MQTTProtocol
        else:
            raise ProfileValueError("profile value not supported" , self.profile)
        
        v = self.queuePublishTx.get(addr, deque())
        self.queuePublishTx[addr] = v
        v = self.windowPublish.get(addr, dict() )
        self.windowPublish[addr] = v
        v = self.windowPubRelease.get(addr, dict() )
        self.windowPubRelease[addr] = v
        v = self.windowPubRx.get(addr, dict())
        self.windowPubRx[addr] = v
        v = self.windowSubscribe.get(addr, dict() )
        self.windowSubscribe[addr] = v
        v = self.windowUnsubscribe.get(addr, dict())
        self.windowUnsubscribe[addr] = v

        # Keeps a persistent reference to the last protocol built
        # This is ok *only* when connecting to a single broker. 
        self.protocol = MQTTProtocol(self, addr)
        return self.protocol


    def clientConnectionLost(self, connector, reason):
        log.warn('Lost connection. Reason {reason!r}:', reason=reason)
        ReconnectingClientFactory.clientConnectionLost(self, connector, reason)


    def clientConnectionFailed(self, connector, reason):
        log.warn('Conenction failed. Reason {reason!r}:', reason=reason)
        ReconnectingClientFactory.clientConnectionFailed(self, connector,
                                                         reason)

    # -------------
    # Helper methods
    # --------------

    def makeId(self):
        '''Produce ids for Protocol packets, outliving their sessions'''
        self.id = (self.id + 1) % 65536
        self.id = self.id or 1   # avoid id 0
        return self.id


__all__ = [MQTTFactory]



from setuptools import setup, Extension, find_packages
import versioneer

# Default description in markdown
long_description = open('README.md').read()
 
# Converts from makrdown to rst using pandoc
# and its python binding.
# Docunetation is uploaded in PyPi when registering
# by issuing `python setup.py register`

try:
    import subprocess
    import pandoc
 
    process = subprocess.Popen(
        ['which pandoc'],
        shell=True,
        stdout=subprocess.PIPE,
        universal_newlines=True
    )
 
    pandoc_path = process.communicate()[0]
    pandoc_path = pandoc_path.strip('\n')
 
    pandoc.core.PANDOC_PATH = pandoc_path
 
    doc = pandoc.Document()
    doc.markdown = long_description
 
    long_description = doc.rst
 
except:
    pass
   


classifiers = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.4',
    'Framework :: Twisted',	
    'Topic :: Communications',
    'Topic :: Internet',
    'Development Status :: 4 - Beta',
]


setup(name             = 'twisted-mqtt',
      version          = versioneer.get_version(),
      cmdclass         = versioneer.get_cmdclass(),
      author           = 'Rafael Gonzalez',
      author_email     = 'astrorafael@yahoo.es',
      description      = 'MQTT client protocol package for Twisted',
      long_description = long_description,
      license          = 'MIT',
      keywords         = 'Python Twisted',
      url              = 'http://github.com/astrorafael/twisted-mqtt/',
      classifiers      = classifiers,
      packages         = find_packages(exclude=['mqtt.test', 'mqtt.client.test']),
      install_requires = ['twisted >= 16.2.0'],
  )
